import argparse
import importlib
from cocotask import CocoConsumerManager
import logging
import json
from jsmin import jsmin
import sys
import codecs


def class_for_name(module_name, class_name, path):
    sys.path.append(path)
    # load the module, will raise ImportError if module cannot be loaded
    m = importlib.import_module(module_name)
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, class_name)
    return c


def load_config(path):
    with codecs.open(path, 'r', 'utf-8') as f:
        config = json.loads(jsmin(f.read()))

    return config


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('config', help = 'define config file path')
    parser.add_argument('module', help = 'define the module name')
    parser.add_argument('worker', help = 'define the worker class name')
    parser.add_argument('worker_number', help = 'number of workers to work', type=int)
    parser.add_argument('--logginglevel', help = 'define logging level')
    parser.add_argument('--modulepath', default = '.', help = 'path to the module. Default to .')

    args = parser.parse_args()

    if args.logginglevel:
        level = getattr(logging, args.logginglevel.upper(), None)
        logging.basicConfig(level=level)

    config = load_config(args.config)
    worker_class = class_for_name(args.module, args.worker, args.modulepath)

    manager = CocoConsumerManager(config, worker_class, args.worker_number)
    manager.start()


if __name__ == '__main__':  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
import logging
import sys
import os


def init(name, logfile, level=logging.DEBUG, with_console=True):
    global logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    formatter = logging.Formatter("[%(levelname)-7s][%(asctime)s][%(filename)s:%(lineno)3d] %(message)s",
                                  "%Y/%m/%d %H:%M:%S")

    if with_console:
        h1 = logging.StreamHandler(sys.stdout)
        h1.setFormatter(formatter)
        logger.addHandler(h1)

    d = os.path.dirname(logfile)
    if d != "" and not os.path.exists(d):
        try:
            os.makedirs(os.path.dirname(logfile))
        except OSError as e:
            logger.warning("Unable to create log file at %s, Exp -> %s" % (logfile, repr(e)))
            return logger

    h2 = logging.FileHandler(logfile)
    h2.setFormatter(formatter)
    logger.addHandler(h2)

    return logger


logger = init(__name__, os.getenv("COCOTASK_LOG", "cocotask.log"))



from multiprocessing import Pool
from .rabbitmq.rmq_consumer import CocoRMQConsumer
from .kafka.kafka_consumer import CocoKafkaConsumer
from .redis.redis_consumer import CocoRedisConsumer
from .logger import logger
import time


# this is aka mixture of a consumer factory and threadpool
class CocoConsumerManager(object):

    CONSUMER_CLASS = {
        "RMQ": CocoRMQConsumer,
        "KAFKA": CocoKafkaConsumer,
        "REDIS": CocoRedisConsumer
    }

    def __init__(self, config, worker_class, pool_size, customized_logger = None):
        self._worker_class = worker_class
        self._pool = None
        self._pool_size = pool_size
        self._config = config
        if customized_logger:
            self._logger = customized_logger

    def start(self):
        self._pool = Pool()
        _ = [self._pool.apply_async(CocoConsumerManager._start_consumer,
                                    args=[x, self._worker_class, self._config]) for x in range(self._pool_size)]
        self._pool.close()
        self._pool.join()
        logger.warning("All progress stopped!")

    @staticmethod
    def _start_consumer(seq, worker_class, config):
        logger.debug('start consumer')
        worker = worker_class(config, seq)

        consumer_type = config["MQ_TYPE"]
        logger.debug('cusumer type: {}'.format(consumer_type))
        consumer_class = CocoConsumerManager.CONSUMER_CLASS[consumer_type]
        sub_config = config[consumer_type]

        while True:
            try:
                consumer = consumer_class(sub_config, worker, logger)
                consumer.connect()
            except Exception as err:
                    logger.error(err)
                    logger.warning("Consumer Error. Reconnect after 10 seconds.")
                    time.sleep(10)




from ..base_consumer import CocoBaseConsumer
from kafka import KafkaConsumer

class CocoKafkaConsumer(CocoBaseConsumer):
    def __init__(self, conf, worker, logger = None):
    	self._consumer = None
    	super().__init__(conf, worker, logger)
        
    def connect(self):
        username = self._config.get("USERNAME", None)
        password = self._config.get("PASSWORD", None)

        if username and password:
            self._consumer = KafkaConsumer(bootstrap_servers=self._config['BOOTSTRAP_SERVERS'],
                                           group_id=self._config['GROUP_ID'],
                                           security_protocol = "SASL_PLAINTEXT",
                                           sasl_mechanism = 'PLAIN',
                                           sasl_plain_username = username,
                                           sasl_plain_password = password,
                                           consumer_timeout_ms=1000,
                                           auto_offset_reset='earliest'
                                           )
        else:
            self._consumer = KafkaConsumer(bootstrap_servers=self._config['BOOTSTRAP_SERVERS'],
                                           group_id=self._config['GROUP_ID'],
                                           consumer_timeout_ms=1000,
                                           auto_offset_reset='earliest')

        self._consumer.subscribe(self._config['TOPIC'])

        while True:
            for message in self._consumer:
            	try:
                    self._logger.debug(message.value)
                    self._worker.process(message.value)
            	except Exception as err:
            		self._logger.error(err)



# -*- coding: utf-8 -*-

import json
from ..base_producer import CocoBaseProducer
from kafka import KafkaProducer

class CocoKafkaProducer(CocoBaseProducer):

    def __init__(self, conf, logger = None):
        self._producer = None
        super().__init__(conf, logger)

    def connect(self):
      username = self._config.get("USERNAME", None)
      password = self._config.get("PASSWORD", None)

      if username and password:
        self._producer = KafkaProducer(bootstrap_servers=self._config['BOOTSTRAP_SERVERS'],
                                       security_protocol = "SASL_PLAINTEXT",
                                       sasl_mechanism = 'PLAIN',
                                       sasl_plain_username = username,
                                       sasl_plain_password = password
                                       )
      else:
        self._producer = KafkaProducer(bootstrap_servers=self._config['BOOTSTRAP_SERVERS'])


    def send(self, data):
      databytes = data
      if isinstance(databytes, str):
        databytes = databytes.encode()

      self._producer.send(self._config['TOPIC'], databytes)

    def close(self):
        self._producer.close()





from abc import ABC, abstractmethod
import logging
default_logger = logging.getLogger(__name__)

class CocoBaseProducer(ABC):
    def __init__(self, conf, logger = None):
        self._config = conf
        self._logger = logger if logger else default_logger

        super().__init__()


    @abstractmethod
    def connect(self):
        pass


    @abstractmethod
    def send(self, data):
        pass


    @abstractmethod
    def close(self):
        pass




# -*- coding: utf-8 -*-
import pika.exceptions as exceptions
from ..base_consumer import CocoBaseConsumer
from .common import createBlockingConnection
from ..logger import logger
import time


class CocoRMQConsumer(CocoBaseConsumer):

    def __init__(self, conf, worker, logger = None):
        self._exchange_name = conf['EXCHANGE_NAME']
        self._queue_name = conf['QUEUE_NAME']
        self._connection = None
        self._t = None
        super().__init__(conf, worker, logger)

    def connect(self):
        logger.info("CocoRMQConsumer starts working")
        while True:
            logger.info("Trying to connect RabbitMQ server ...")
            try:
                self._connection, channel = createBlockingConnection(self._config)
                break
            except exceptions.AMQPConnectionError as e:
                logger.error("Connect failed, exp: %s" % e)
                logger.debug("Will Try later after 10 seconds...")
                time.sleep(10)

        logger.info("Connected, starts consuming ...")
        channel.queue_bind(exchange=self._exchange_name, queue=self._queue_name)
        channel.basic_qos(prefetch_count=1)
        channel.basic_consume(self.on_message, queue=self._queue_name, no_ack=True)
        try:
            channel.start_consuming()
        except exceptions.AMQPConnectionError or exceptions.ChannelError as e:
            logger.error("AMQPConnection or Channel error, exp: %s" % e)
            logger.debug("Trying to reconnect ...")
            self.connect()

        logger.error("Consuming stopped")

    def on_message(self, ch, method, properties, body):
        self._worker.process(body)
